package lsedit;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.*;

import java.util.*;
import java.io.*;
import javax.swing.*;

import java.awt.geom.*;

public class RelationInstance extends LandscapeObject /*  extends JComponent */ implements MouseListener, MouseMotionListener  
{
	class DrawInfo extends Object {
		public void		getBounds(Rectangle r)
		{
			r.setBounds(0,0,0,0);
		}
		
		public boolean contains(int x, int y)	// (x,y) w.r.t diagram
		{
			return false;
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
		}

		public boolean overEdgePoint(int x, int y)
		{
			return false;
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			return null;
		}
	}

	class ArcInfo extends DrawInfo {

		int	m_x;	// Top left hand corner of arc bounded area
		int m_y;

		/* arrow head at end of arc

             (x1, y1)
				|     \
				|       \
                |         \
				|           \
		(x1, y1+radius-1)-- (x1+radius-1, y1+radius-1)
		*/

		public ArcInfo()
		{
		}

		public void setArc(int x, int y)
		{
			m_x = x;
			m_y = y;
		}

		public void getBounds(Rectangle r)
		{
			r.setBounds(m_x, m_y, RelationInstance.RECURSIVE_LOOP_DIAMETER, RelationInstance.RECURSIVE_LOOP_DIAMETER);
		}

		public boolean contains(int x, int y)		// (x,y) w.r.t. diagram
		{
			int x1, y1;
			int	r;

			// Transform axis so that over centre of arc
		
			x1 = x - m_x - RelationInstance.RECURSIVE_LOOP_RADIUS;
			y1 = y - m_y - RelationInstance.RECURSIVE_LOOP_RADIUS;

			if (x1 >= 0 || y1 <= 0) {
				// We never paint inside the object the arc is attached to so anything in this quadrant is false
				r = (x1*x1 + y1*y1);

				// Are we within one pixel either way of lying on the loop that forms the arc
				if (r >= ((RelationInstance.RECURSIVE_LOOP_RADIUS - 1) * (RelationInstance.RECURSIVE_LOOP_RADIUS - RelationInstance.NEAR_LOOP_THRESHOLD)) && r <= ((RECURSIVE_LOOP_RADIUS+1)*(RECURSIVE_LOOP_RADIUS+RelationInstance.NEAR_LOOP_THRESHOLD))) {
					return(true);
			}	}
			return(false);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			g.drawArc(0, 0, RelationInstance.RECURSIVE_LOOP_DIAMETER, RelationInstance.RECURSIVE_LOOP_DIAMETER, 270, 270);
			g.fillPolygon(RelationInstance.m_arc_xp, RelationInstance.m_arc_yp, 3);
		}
	}	

	class LineInfo extends DrawInfo {

		int m_style;

		public void setStyle(int style)
		{
			m_style = style;
		} 
	}

	class ElidedInfo extends LineInfo {

		public void getBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
		{
			RelationInstance.getLineBounds(srcX, srcY, dstX, dstY, r);
			RelationInstance.includeArrowBounds(srcX, srcY, dstX, dstY, r);
		}

		public void paintComponent(Graphics g, int srcX, int srcY, int dstX, int dstY)
		{
			Util.drawSegment(g, m_style, srcX, srcY, dstX, dstY);
			Util.drawArrowHead(g, srcX, srcY, dstX, dstY);
		}
	}

	class OutElidedInfo extends ElidedInfo {

		private EdgePoint m_src;
		int		m_dstX, m_dstY;

		public void setOutElidedInfo(EdgePoint srcPt, Rectangle dstLyt)
		{
			m_src  = srcPt;

			double    dstX  = (double) (dstLyt.x + dstLyt.width/2);		// Centre of destination
			double	  dstY  = (double) (dstLyt.y + dstLyt.height/2);

			double dx = srcPt.getX() - dstX;
			double dy = srcPt.getY() - dstY;
			double df = Math.sqrt(dx*dx + dy*dy);

			// Make line length twice arrow length

			m_src  = srcPt;
			m_dstX = (int) (srcPt.getX() - dx*2*Util.ARROW_L/df);
			m_dstY = (int) (srcPt.getY() - dy*2*Util.ARROW_L/df);
		}

		public void getBounds(Rectangle r)
		{
			getBounds((int) m_src.getX(), (int) m_src.getY(), m_dstX, m_dstY, r);
		}

		// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

		public boolean contains(int x, int y) // (x,y) w.r.t diagram
		{
			return(RelationInstance.nearLine((int) m_src.getX(), (int) m_src.getY(), m_dstX, m_dstY, x, y));
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			if (RelationInstance.nearPoint((int) m_src.getX(), (int) m_src.getY(), x, y)) {
				return(m_src);
			}
			return(null);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			paintComponent(g, (int) (m_src.getX() - shiftX), (int) (m_src.getY() - shiftY), m_dstX  - shiftX, m_dstY  - shiftY);
		}
	}

	class InElidedInfo extends ElidedInfo {

		int		m_srcX, m_srcY;
		private EdgePoint m_dst;

		public void setInElidedInfo(Rectangle srcLyt, EdgePoint dstPt)
		{
			double    srcX  = (double) (srcLyt.x + srcLyt.width/2);
			double	  srcY  = (double) (srcLyt.y + srcLyt.height/2);

			double dx = srcX - dstPt.getX();
			double dy = srcY - dstPt.getY();
			double df = Math.sqrt(dx*dx + dy*dy);
			   
			// Make line length twice arrow length

			m_srcX = (int) (dstPt.getX() + (dx*2*Util.ARROW_L/df));
			m_srcY = (int) (dstPt.getY() + (dy*2*Util.ARROW_L/df));
			m_dst  = dstPt;
		}

		public void getBounds(Rectangle r)
		{
			getBounds(m_srcX, m_srcY, (int) m_dst.getX(), (int) m_dst.getY(), r);
		}

		// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

		public boolean contains(int x, int y) // (x,y) w.r.t diagram
		{
			return(RelationInstance.nearLine(m_srcX, m_srcY, (int) m_dst.getX(), (int) m_dst.getY(), x, y));
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			if (RelationInstance.nearPoint((int) m_dst.getX(), (int) m_dst.getY(), x, y)) {
				return(m_dst);
			}
			return(null);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			paintComponent(g, m_srcX - shiftX, m_srcY - shiftY, (int) (m_dst.getX()  - shiftX), (int) (m_dst.getY()  - shiftY));
		}
	}

	class EdgeInfo extends LineInfo {

		private int		  m_embellished;

		private EdgePoint m_src;
		private EdgePoint m_dst;

		public void setEdgeInfo(EdgePoint src, EdgePoint dst)
		{
/*
			if (src == null) {
				System.out.println("Source edgeInfo is null");
			}
 			if (dst == null) {
				System.out.println("Destination edgeInfo is null");
			}
 */
			m_src  = src;
			m_dst  = dst;
		}

		public void setEmbellished(int embellished)
		{
			m_embellished = embellished;
		}

		public void getBounds(Rectangle r)
		{
			RelationInstance.getLineBounds((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), r);

			if ((m_embellished & DRAW_TAIL_MARK) != 0) {
				RelationInstance.includeTailBounds((int) m_src.getX(), (int) m_src.getY(), r);
			}
			if ((m_embellished & DRAW_CENTRE_MARK) != 0) {
				RelationInstance.includeCentreBounds((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), r);
			}
			if ((m_embellished & DRAW_ARROW_MARK) != 0) {
				RelationInstance.includeArrowBounds((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), r);
			}
		}

		// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

		public boolean contains(int x, int y) // (x,y) w.r.t diagram
		{
			return(RelationInstance.nearLine((int) m_src.getX(), (int) m_src.getY(), (int) m_dst.getX(), (int) m_dst.getY(), x, y));
		}

		public EdgePoint mouseOverEdgePoint(int x, int y)
		{
			if (RelationInstance.nearPoint((int) m_src.getX(), (int) m_src.getY(), x, y)) {
				return(m_src);
			}
			if (RelationInstance.nearPoint((int) m_dst.getX(), (int) m_dst.getY(), x, y)) {
				return(m_dst);
			}
			return(null);
		}

		public void paintComponent(Graphics g, int shiftX, int shiftY)
		{
			int srcX = (int) (m_src.getX() - shiftX);
			int srcY = (int) (m_src.getY() - shiftY);
			int dstX = (int) (m_dst.getX() - shiftX);
			int dstY = (int) (m_dst.getY() - shiftY);

/*			if (m_src.getEntity().m_label.equals("code")) {
				System.out.println("EdgeInfo.paintComponent srcY=" + m_src.getY());
			}
*/
			
			Util.drawSegment(g, m_style, srcX, srcY, dstX, dstY);
			if ((m_embellished & DRAW_TAIL_MARK) != 0) {
				g.fillOval(srcX-RelationInstance.NEAR_PIXEL_SIZE/2, srcY-RelationInstance.NEAR_PIXEL_SIZE/2, RelationInstance.NEAR_PIXEL_SIZE, RelationInstance.NEAR_PIXEL_SIZE);
			}
			if ((m_embellished & DRAW_CENTRE_MARK) != 0) {
				Color tc = g.getColor();
			
				g.setColor(HIGHLIGHT_COLOUR);
				g.fillRect((srcX + dstX - HIGHLIGHT_DIM)/2, (srcY + dstY - HIGHLIGHT_DIM)/2, HIGHLIGHT_DIM, HIGHLIGHT_DIM);
				g.setColor(tc);
			}
			if ((m_embellished & DRAW_ARROW_MARK) != 0) {
				Util.drawArrowHead(g, srcX, srcY, dstX, dstY);
			}
		}
	}

	// Constants

	public    final static int    CLIENT_SUPPLIER_EL_LEN = 16;
	protected final static double SEP_THRESHOLD          = 10.0;
	protected final static Color  HIGHLIGHT_COLOUR       = Color.yellow;
	protected final static int	  HIGHLIGHT_DIM	         = 4;
	protected final static int	  NEAR_PIXEL_SIZE		 = 6;

	// First order attribute ids

	protected final static String STYLE_ID	= "style";

	// Bits set in m_marks

	protected final static int LOOP_MARK	    	= 0x01;		// Repaint this edge
	protected final static int NORMAL_MARK			= 0x02;
	protected final static int ELISION_MARK         = 0x04;
	protected final static int OUT_ELIDED_MARK		= 0x10;
	protected final static int IN_ELIDED_MARK		= 0x20;
	protected final static int IN_DIAGRAM_MARK		= 0x40;

	protected final static int VALID_MARK           = 0x100;

	public final static int GROUP_FLAG_MARK         = 0x200;
	public final static int HIGHLIGHT_FLAG_MARK     = 0x400;
	public final static int	ADD_RESULTS_MARK		= 0x800;

	public final static int PRESENTATION_MARKS      = GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK;

	// Global values for relations

	// Static methods must be in the top level class
	
	protected final static int	RECURSIVE_LOOP_RADIUS	 = 8;
	protected final static int	RECURSIVE_LOOP_DIAMETER  = RECURSIVE_LOOP_RADIUS * 2;

	protected final static int[]  m_arc_xp = {0, 0, RECURSIVE_LOOP_RADIUS-1};
	protected final static int[]  m_arc_yp = {RECURSIVE_LOOP_RADIUS-1, 0, RECURSIVE_LOOP_RADIUS-1};
	
	protected final static int NEAR_POINT_THRESHOLD = 1;
	protected final static int NEAR_LINE_THRESHOLD2 = 1;		// Square of threshold to avoid using sqrt
	protected final static int NEAR_LOOP_THRESHOLD  = 1;

	public final static int DRAW_TAIL_MARK	 = 0x01;
	public final static int DRAW_CENTRE_MARK = 0x02;
	public final static int DRAW_ARROW_MARK  = 0x04;
		
	// Properties

	// An edge is defined by a relation class and two entity instances

	protected RelationClass	  parentClass;				// The class this relation belongs to
	protected EntityInstance  m_src,     m_dst;			// The real source and destination of this edge
	protected EntityInstance  m_drawSrc, m_drawDst;		// The derived source and destination of this edge given closed entities exist

	// Variables

	protected int			  m_mark = 0;				// A collection of bits

	protected Object		  m_drawInfo;

	// First order attributes

	protected int			  timeStamp = -1;

	// Stupidity: All these methods have to go in RelationInstance because java doesn't allow static inner functions

	public static void getLineBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		if (srcX <= dstX) {
			x1 = srcX;
			x2 = dstX;
		} else {
			x1 = dstX;
			x2 = srcX;
		}
		if (srcY <= dstY) {
			y1 = srcY;
			y2 = dstY;
		} else {
			y1 = dstY;
			y2 = srcY;
		}
		if (Util.LINE_W > 1) {
			int w, h, shift;

			w     = x2 - x1;
			h     = y2 - y1;
			shift = Util.LINE_W/2;

				
			if (h > 2*w) {
				x1 -= shift;
				x2 += shift;
			}
			if (w >= 2*h) {
				y1 -= shift;
				y2 += shift;
			}
		}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}
	
	public static void includeTailBounds(int srcX, int srcY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		x1 = r.x;
		y1 = r.y;
		x2 = x1 + r.width;
		y2 = y1 + r.height;

		x = srcX - RelationInstance.NEAR_PIXEL_SIZE/2;
		y = srcY - RelationInstance.NEAR_PIXEL_SIZE/2;

		if (x < x1) {
			x1 = x;
		}
		if (y < y1) {
			y1 = y;
		} 
		x += RelationInstance.NEAR_PIXEL_SIZE;
		y += RelationInstance.NEAR_PIXEL_SIZE;

		if (x > x2) {
			x2 = x;
		}
		if (y > y2) {
			y2 = y;
		}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}

	public static void includeCentreBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		x1 = r.x;
		y1 = r.y;
		x2 = x1 + r.width;
		y2 = y1 + r.height;

		x = (srcX + dstX - HIGHLIGHT_DIM)/2;
		y = (srcY + dstY - HIGHLIGHT_DIM)/2;

		if (x < x1) {
			x1 = x;
		}
		if (y < y1) {
			y1 = y;
		}
		x += HIGHLIGHT_DIM;
		y += HIGHLIGHT_DIM;
		if (x > x2) {
			x2 = x;
		}
		if (y > y2) {
			y2 = y;
		}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}

	public static void includeArrowBounds(int srcX, int srcY, int dstX, int dstY, Rectangle r)
	{
		int x1, y1, x2, y2, x, y;

		x1 = r.x;
		y1 = r.y;
		x2 = x1 + r.width;
		y2 = y1 + r.height;

		int[] x3 = new int[3];
		int[] y3 = new int[3];

		if (Util.getArrow(srcX, srcY, dstX, dstY, x3, y3)) {
			int i;

			for (i = 0; i < 3; ++i) {
				x = x3[i];
				y = y3[i];

				if (x < x1) {
					x1 = x;
				}
				if (y < y1) {
					y1 = y;
				}
				if (x > x2) {
					x2 = x;
				}
				if (y > y2) {
					y2 = y;
		}	}	}
		r.setBounds(x1, y1, x2-x1, y2-y1);
	}
		
	// Test to see if (x,y) is near line from (x1,y1) to (x2,y2)

	public static boolean nearPoint(int srcX, int srcY, int x, int y) 
	{
		int diff;

		diff = x-srcX;
		if (diff < 0) {
			diff = -diff;
		}
		if (diff > RelationInstance.NEAR_POINT_THRESHOLD) {
			return(false);
		}
		diff = y-srcY;
		if (diff < 0) {
			diff = -diff;
		}
		if (diff > RelationInstance.NEAR_POINT_THRESHOLD) {
			return(false);
		}
		return(true);
	} 

	public static boolean nearLine(int srcX, int srcY, int dstX, int dstY, int x, int y) 
	{
		if (srcX == dstX && srcY == dstY) {
			// Degenerate case line becomes a point
			return(RelationInstance.nearPoint(srcX, srcY, x, y));
		}
		
		// Determine distance to line and point R(x, y)
		//
		// Using parametric form for equation of line
		// L = P0 + tV	where V is vector from P0 to P1
		//
		// Find t such that vector R -> P0 + tV is orthogonal to V.
		// We then restrict to checking distances to
		// segment P0, P1 (0 < t < 1). 

		double rx = x;
		double ry = y;
		double vx = dstX - srcX;
		double vy = dstY - srcY; 
		
		double t = ((rx - srcX) * vx + (ry - srcY) * vy) / (vx*vx + vy*vy);

		if (t >= 0 && t <= 1.0) {
			// Only check for segment between P0 and P1
			double px = srcX + t * vx;
			double py = srcY + t * vy; 
			double lx = (rx - px);
			double ly = (ry - py);

			double dist2 = lx*lx + ly*ly;

			if (dist2 <= NEAR_LINE_THRESHOLD2) {
				return(true);
		}	}	
		return(false);
	}	
						
	protected boolean processFirstOrderAttributes(Attribute attr, Hashtable entities) 
	{
		boolean hasVal = (attr.avi != null);

		if (attr.id.equals(COLOUR_ID)) {
			if (hasVal)
				setObjectColor(attr.parseColour());
			return true; 
		}
		if (attr.id.equals(STYLE_ID)) {
			if (hasVal) {
				setStyle(attr.parseInt());
			}
			return true;
		}
		return false;
	}

	protected Vector points(boolean allowElision) {

		EntityInstance	root      = dg.getDrawRoot();
		EntityInstance	src       = m_drawSrc;
		EntityInstance	dst       = m_drawDst;
		OutElidedInfo	outElided = null;
		InElidedInfo	inElided  = null;
		EdgeInfo		edgeInfo;
		int				style;
		int				embellished;

		// We will travel from the source to the destination
		// going first up then down using as our points of traversal the inflection points

		EntityInstance endUp, endDown;

		boolean client   = !root.hasDescendantOrSelf(src);
		boolean supplier = !root.hasDescendantOrSelf(dst);

		if (!client && !supplier) {
			endUp   = src.commonAncestor(dst); // Common container
			endDown = endUp;
		} else if (client) {
			endUp   = src.m_containedBy;
			endDown = root.m_containedBy;
		} else {
			endUp   = root.m_containedBy;
			endDown = dst.m_containedBy;
		}

		if (endUp == null) {
			System.out.println("C Error: " + m_src + " " + m_dst + " " + client + " " + supplier);
			return null;
		}

		if (endDown == null) {
			System.out.println("S Error: " + m_src + " " + m_dst + " " + client + " " + supplier);
			return null;
		}

		// Not in the same container

		Vector v1 = new Vector();
		Vector v2 = new Vector();

		EntityInstance e, sp, dp, ep;
	    EdgePoint pt1, pt2, ppt;

		for (sp = m_drawSrc; sp.m_containedBy != endUp; ) {
			sp = sp.m_containedBy;
		}
		for (dp = m_drawDst; dp.m_containedBy != endDown; ) {
			dp = dp.m_containedBy;
		}
		Rectangle spLyt = sp.getDiagramBounds();
		Rectangle dpLyt = dp.getDiagramBounds();

		if (allowElision && (m_src != m_drawSrc || m_dst != m_drawDst)) {
			style = Util.LINE_STYLE_DOTTED;
		} else {
			style = getStyle();
		}

		embellished = RelationInstance.DRAW_TAIL_MARK;
		if (isMarked(GROUP_FLAG_MARK)) {
			embellished |= RelationInstance.DRAW_CENTRE_MARK;
		}

		/* The following is a bit tricky.
		 *
		 * We build up two vectors of points which represent the
		 * the curve we want the edge to take. These sets are built
		 * one point at a time. 
		 *
		 * The first set builds a curve from the source entity 
		 * to the common container entity. The second set builds
		 * the curve from the destination to the common container.
		 * The second curve then has its points reversed an is 
		 * appended to the first curve. 
		 *
		 * Client/Supplier edge elision complicated things because
		 * it requires the ability to have discontinuous curves.
		 *
		 */

		ppt = pt1 = null;		// Previous point
		for (e = src; e != endUp; e = ep) {
			if (e == null) {
				MsgOut.println("Error in points: " + "(" + parentClass.getId() + " " +  m_src.getId() + " " + m_dst.getId() + ")");
				return null;
			}

			ep  = e.m_containedBy;
			pt1 = e.getOutPoint(this, getEdgeMode(), spLyt, dpLyt);

			if (ppt != null) {
				if (allowElision && e.isExitingRelationElided(parentClass)) {
					outElided = new OutElidedInfo();
					outElided.setOutElidedInfo(pt1, dpLyt);
					outElided.setStyle(style);
					v1.addElement(outElided);
					break;
				}
				edgeInfo = new EdgeInfo();
				edgeInfo.setEdgeInfo(ppt, pt1);
				edgeInfo.setStyle(style);
				edgeInfo.setEmbellished(embellished);
				embellished &= ~DRAW_TAIL_MARK;
				v1.addElement(edgeInfo);
			}
			ppt = pt1;
		}

		// We've found all points from source entity to 
		// the edge of its ancestor in the common container
	
		// Finally, find the points from the destination entity
		// to the edge of its ancestor in the common container.
		// We'll reverse the order later.

		embellished &= ~RelationInstance.DRAW_TAIL_MARK;
		embellished |= RelationInstance.DRAW_ARROW_MARK;

		ppt = pt2 = null;
		for (e = dst; e != endDown; e = ep) {
			ep  = e.m_containedBy;
			pt2 = e.getOutPoint(this, getEdgeMode(), dpLyt, spLyt);

			if (ppt != null) {
				if (allowElision && e.isEnteringRelationElided(parentClass)) {
					inElided = new InElidedInfo();
					inElided.setInElidedInfo(spLyt, pt2);
					inElided.setStyle(style);
					v2.addElement(inElided);
					break;
				}
				edgeInfo = new EdgeInfo();
				edgeInfo.setEdgeInfo(pt2, ppt);
				edgeInfo.setStyle(style);
				edgeInfo.setEmbellished(embellished);
				embellished &= ~DRAW_ARROW_MARK;
				v2.addElement(edgeInfo);
			}
			ppt = pt2;
		} 

		if (outElided == null && inElided == null) {
			edgeInfo = new EdgeInfo();
			edgeInfo.setEdgeInfo(pt1, pt2);
			v1.addElement(edgeInfo);
		} else if (inElided == null) {
			v2.removeAllElements();
		} else if (outElided == null) {
			v1.removeAllElements();
		}

		for (int i = v2.size()-1; i >= 0; i--) {
			v1.addElement(v2.elementAt(i));
		}
		return v1;
	}

	protected int findNearestSeg(int x, int y, int threshold, Vector pts) 
	{
		double		threshold2;
		Enumeration en = pts.elements();
		RealPoint srcPt, dstPt;
		int ind = 0;

		threshold2 = ((double) threshold) * ((double) threshold);
		dstPt = (RealPoint) en.nextElement();

		do {
			srcPt = dstPt;
			dstPt = (RealPoint) en.nextElement();

			if (!srcPt.elision && srcPt.active) {

				// Determine distance to line and point R(x, y)
				//
				// Using parametric form for equation of line
				// L = P0 + tV	where V is vector from P0 to P1
				//
				// Find t such that vector R -> P0 + tV is orthogonal to V.
				// We then restrict to checking distances to
				// segment P0, P1 (0 < t < 1). 

				double rx = x;
				double ry = y;

				double vx = dstPt.getX() - srcPt.getX();
				double vy = dstPt.getY() - srcPt.getY(); 

				double t = ((rx - srcPt.getX()) * vx + (ry - srcPt.getY()) * vy) / (vx*vx + vy*vy);

				if (t > 0 && t < 1.0) {
					// Only check for segment between P0 and P1
					double px = srcPt.getX() + t * vx;
					double py = srcPt.getY() + t * vy; 

					double lx = (rx - px);
					double ly = (ry - py);

					double dist2 = lx*lx + ly*ly;

					if (dist2 < threshold2) {
//						System.out.println("RelationInstance.findNearestSeg returned " + ind);
						return ind;
				}	}
			}

			ind++;

		} while (en.hasMoreElements());

		return -1;
	}

	// --------------
	// Public methods
	// --------------

	public RelationInstance(RelationClass parentClass, EntityInstance src, EntityInstance dst, Diagram dg)
	{
		this.parentClass = parentClass;
		this.m_src       = src;
		this.m_dst       = dst;
		this.m_drawSrc   = src;
		this.m_drawDst   = dst;
		this.dg	         = dg;

		setObjectColor(parentClass.getObjectColor());
		addMouseListener(this);
		addMouseMotionListener(this);
	}

	public void clearMark(int preserve_relation_marks)
	{
		m_mark    &= preserve_relation_marks;
		m_drawSrc  = null;
		m_drawDst  = null;
		m_drawInfo = null;

//		System.out.println("RelationInstance.clearMark " + this);
	}

	public void orMark(int value)
	{
		m_mark |= value;
	}

	public void nandMark(int value)
	{
		m_mark &= ~value;
	}

	public boolean isMarked(int value)
	{
		return((m_mark & value) != 0);
	}

	public boolean isSrcVisible()
	{
		return (m_src == m_drawSrc);
	}

	public boolean isDstVisible()
	{
		return (m_dst == m_drawDst);
	}


	public LandscapeObject derivedFrom(int i)
	{
		return((i == 0) ? parentClass : null);
	}

	public void computeBounds(Rectangle r)
	{
		DrawInfo drawInfo;

		if (m_drawInfo == null) {
			r.setBounds(0,0,0,0);
		} else if (m_drawInfo instanceof DrawInfo) {
			((DrawInfo) m_drawInfo).getBounds(r);
		} else {
			Enumeration en;
			int x1,y1,x2,y2, x, y;

			x1 = Integer.MAX_VALUE;
			x2 = Integer.MIN_VALUE;
			y1 = Integer.MAX_VALUE;
			y2 = Integer.MIN_VALUE;

			for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
				drawInfo = (DrawInfo) en.nextElement();
				drawInfo.getBounds(r);

				x = r.x;
				if (x < x1) {
					x1 = x;
				}
				y = r.y;
				if (y < y1) {
					y1 = y;
				}
				x += r.width;
				y += r.height;
				if (x > x2) {
					x2 = x;
				}
				if (y > y2) {
					y2 = y;
			}	}
			r.setBounds(x1,y1, x2-x1, y2-1);
	}	}

	public boolean contains(int x, int y)	// (x,y) w.r.t. my bounds
	{
//		System.out.println("RelationInstance.contains " + this);
		if (super.contains(x, y)) {
			// Translate so w.r.t diagram
			x += getX();
			y += getY();
			if (m_drawInfo != null) {
				if (m_drawInfo instanceof DrawInfo) {
					return ((DrawInfo) m_drawInfo).contains(x, y);
				}

				Enumeration en;

				for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
					if ( ((DrawInfo) en.nextElement()).contains(x, y)) {
						return(true);
		}	}	}	}
		return(false);
	}

	public EdgePoint mouseOverEdgePoint(int x, int y)
	{
		if (m_drawInfo != null) {
			if (m_drawInfo instanceof DrawInfo) {
				return ((DrawInfo) m_drawInfo).mouseOverEdgePoint(x, y);
			}

			Enumeration en;
			EdgePoint	ep;

			for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
				ep = ((DrawInfo) en.nextElement()).mouseOverEdgePoint(x,y);
				if (ep != null) {
					return(ep);
		}	}	}
		return(null);
	}

	/* Mark the edge as invalid so we paint it only once and then mark valid
	   -- not twice once from the source and once from the destination
	*/

	public void invalidateEdge()
	{
		nandMark(VALID_MARK);
	}

	// Compute the rules for how an edge is to be drawn between two entities, handling elision settings.
	// Essentially just flag the relation appropriately to assist the verify process know what it is dealing with

	public void draw(boolean allowElision) 
	{
		int				marks;
		boolean			src_elided, dst_elided;

		if (isMarked(VALID_MARK)) {
			return;
		}

	/*	System.out.println("RelationInstance:draw(" + allowElision + ") " + this);
		java.lang.Thread.dumpStack();
		System.out.println("-----");
	*/
		marks = 0;
		if (!allowElision || parentClass.isClassVisible()) {
			if (m_src == null || m_dst == null) {
				MsgOut.println("Landscape error: null src or dst");
				return;
			}
			if (m_src.m_containedBy == null) {
				MsgOut.println("drawEdge: containedBy error");
				return;
			}
			if (m_dst.m_containedBy == null) {
				MsgOut.println("drawEdge: containedBy error");
				return;
			}
					
			EntityInstance	src, dst, ancestor, drawRoot;

			src      = m_drawSrc;
			dst      = m_drawDst;

			if (src == null) {
				System.out.println("draw() " + this + " has no draw src");
				return;
			}
			if (dst == null) {
				System.out.println("draw() " + this + " has no draw dst");
				return;
			}

			drawRoot = dg.getDrawRoot();
			if (drawRoot.hasDescendantOrSelf(src) || drawRoot.hasDescendantOrSelf(dst)) {
				// At least one edge addresses something in the diagram

				if (src != dst || src == m_src || dst == m_dst) {
					// Don't show internal edges beneath closed objects period
					// Such edges will have the same draw src and destination and both real edges lifted to this common closed ancestor

					marks = NORMAL_MARK;
					if (allowElision) {
						String			id        = parentClass.getId();
						EntityInstance	srcParent = src.getContainedBy();
						EntityInstance	dstParent = dst.getContainedBy();

						if (srcParent != null && srcParent == dstParent && srcParent.isInternalRelationElided(id)) {
							// Don't show -- this internal relation is elided
							marks = 0;
						} else {
							if (src.isSrcRelationElided(id)) {
								marks |= OUT_ELIDED_MARK;
							} else {
								for (; srcParent != null && !srcParent.hasDescendantOrSelf(dst); srcParent = srcParent.getContainedBy()) {
									// Source goes from inside srcParent to outside of it
									if (srcParent.isExitingRelationElided(id)) {
										marks |= OUT_ELIDED_MARK;
										break;
							}	}	}

							if (dst.isDstRelationElided(id)) {
								marks |= IN_ELIDED_MARK;
							} else {
								for (; dstParent != null && !dstParent.hasDescendantOrSelf(src); dstParent = dstParent.getContainedBy()) {
									// Destination is inside while src is outside
									if (dstParent.isEnteringRelationElided(id)) {
										marks |= IN_ELIDED_MARK;
										break;
					}	}	}	}	}

					if (marks == NORMAL_MARK && src == dst) {
						// Recursive loop
						marks = LOOP_MARK;
		}	}	}	}
			
		if (marks == 0) {
			if (isMarked(IN_DIAGRAM_MARK)) {
				dg.remove(this);
				nandMark(IN_DIAGRAM_MARK);
			}
		} else {
			marks |= VALID_MARK;
			if (allowElision) {
				marks |= ELISION_MARK;
			}
			if (!isMarked(IN_DIAGRAM_MARK)) {
				marks |= IN_DIAGRAM_MARK;
				orMark(marks);
				dg.add(this /* , JLayeredPane.PALETTE_LAYER */);	// Add this relation into the diagram
				this.setVisible(true);
				validate();
			} else {
				orMark(marks);
				validate();
	}	}	}

	public void drawHighlighted() 
	{
		if (getHighlightFlag() && getRelationClass().isActive()) {
			draw(false);
	}	}

	// Compute the coordinates for the edge

	public void validate()
	{
		EntityInstance src, dst;
		Rectangle	   r;
		Enumeration		en;
		
//		System.out.println("RelationInstance.validate " + this);
/*		java.lang.Thread.dumpStack();
*/
		super.validate();
		
		src = m_drawSrc;

		if (src == null) {
			// We may validate a relation because we have changed its flags, not realising it is no longer in the diagram
			return;
		}

		dst = m_drawDst;
		if (dst == null) {
			// We may validate a relation because we have changed its flags, not realising it is no longer in the diagram
			return;
		}

		if (isMarked(LOOP_MARK)) {
			// This is a recursive loop from something to itself
			if (m_drawInfo == null || !(m_drawInfo instanceof ArcInfo)) {
				m_drawInfo = new ArcInfo();
			}
//			System.out.println("RelationInstance:validate " + this + " recursive");
			((ArcInfo) m_drawInfo).setArc(src.getDiagramX() + src.getWidth() - RECURSIVE_LOOP_RADIUS, src.getDiagramY() - RECURSIVE_LOOP_RADIUS);
		} else	{
			Rectangle srcLyt    = src.getDiagramBounds();
			Rectangle dstLyt    = dst.getDiagramBounds();

			if (isMarked(OUT_ELIDED_MARK | IN_ELIDED_MARK)) {
				OutElidedInfo out = null;
				InElidedInfo  in  = null;
				DrawInfo	  drawInfo;

				if (isMarked(OUT_ELIDED_MARK)) {
					if (m_drawInfo != null) {
						if (m_drawInfo instanceof OutElidedInfo) {
							out = (OutElidedInfo) m_drawInfo;
						} else if (m_drawInfo instanceof Vector) {
							for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
								drawInfo = (DrawInfo) en.nextElement();
								if (drawInfo instanceof OutElidedInfo) {
									out = (OutElidedInfo) drawInfo;
									break;
					}	}	}	}
					if (out == null) {
						out = new OutElidedInfo();
					}
					// Show that the out edge is elided

					EdgePoint srcPt = src.getOutPoint(this, getEdgeMode(), srcLyt, dstLyt);

					out.setOutElidedInfo(srcPt, dstLyt);
					out.setStyle(getStyle());
//					System.out.println("RelationInstance:validate " + this + " out elided");
				} 

				if (isMarked(IN_ELIDED_MARK)) {
					if (m_drawInfo != null) {
						if (m_drawInfo instanceof InElidedInfo) {
							in = (InElidedInfo) m_drawInfo;
						} else if (m_drawInfo instanceof Vector) {
							for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
								drawInfo = (DrawInfo) en.nextElement();
								if (drawInfo instanceof InElidedInfo) {
									in = (InElidedInfo) drawInfo;
									break;
					}	}	}	}
					if (in == null) {
						in = new InElidedInfo();
					}
										// Show that the in edge is elided
					EdgePoint dstPt = dst.getOutPoint(this, getEdgeMode(), dstLyt, srcLyt); 
	   
					in.setInElidedInfo(srcLyt, dstPt);
					in.setStyle(getStyle());
//					System.out.println("RelationInstance:validate " + this + " in elided");
				}

				if (out == null) {
					m_drawInfo = in;
				} else if (in == null) {
					m_drawInfo = out;
				} else {
					if (m_drawInfo instanceof Vector) {
						((Vector) m_drawInfo).removeAllElements();
					} else {
						m_drawInfo = new Vector();
					}
					((Vector) m_drawInfo).addElement(out);
					((Vector) m_drawInfo).addElement(in);
				}
			} else if (isMarked(NORMAL_MARK)) {
				int	embellished;

				if (src.m_containedBy == dst.m_containedBy || (getEdgeMode() == Do.DIRECT_EDGE_STATE || !isMarked(ELISION_MARK))) {

					// We will draw a direct edge between src and dst

					if (m_drawInfo == null || !(m_drawInfo instanceof EdgeInfo)) {
						m_drawInfo = new EdgeInfo();
					}

					EdgeInfo edgeInfo = (EdgeInfo) m_drawInfo;
					EdgePoint pt1     = src.getOutPoint(this, getEdgeMode(), srcLyt /* src.diagramBounds */, dstLyt /* dst.diagramBounds */);
					EdgePoint pt2     = dst.getOutPoint(this, getEdgeMode(), dstLyt /* dst.diagramBounds */, srcLyt /* src.diagramBounds */);
					edgeInfo.setEdgeInfo(pt1, pt2);

/*
					if (src.getLabel().equals("code")) {
						System.out.println("RelationInstance code source point " + pt1);
					}
 */
					if (m_src != m_drawSrc || m_dst != m_drawDst) {
						edgeInfo.setStyle(Util.LINE_STYLE_DOTTED);
					} else {
						edgeInfo.setStyle(getStyle());
					}
					embellished = RelationInstance.DRAW_ARROW_MARK;
					if (isMarked(GROUP_FLAG_MARK)) {
						embellished |= RelationInstance.DRAW_CENTRE_MARK;
					}
					edgeInfo.setEmbellished(embellished);
				} else {
					m_drawInfo = points(isMarked(ELISION_MARK));
//					System.out.println("RelationInstance:validate " + this + " normal" + m_points);
		}	}	}
		if (m_drawInfo != null) {
			r = new Rectangle();
			computeBounds(r);
			// We don't draw the right and bottom edge
			r.width  += 1;
			r.height += 1;
			setBounds(r);
//			System.out.println("RelationInstance.Bounds " + this + "=" + getBounds());
			repaint();
		}
	}

	// Paint the edge

	public void paintComponent(Graphics g)
	{
		int	shiftX, shiftY, x1, y1, x2, y2;

		shiftX = getX();
		shiftY = getY();
			
		g.setColor(getObjectColor());

		/*		For debugging */
		/*
			System.out.println("RelationInstance.paintComponent " + this + " {" + getX() + "," + getY() + "x" + getWidth() + "," + getHeight() + "}");
			g.drawRect(0, 0, getWidth()-1, getHeight()-1);

		if (m_src.getLabel().equals("code")) {
			System.out.println("RelationInstance.paintComponent " + getBounds());
		}
		*/

		if (m_drawInfo != null) {

			if (m_drawInfo instanceof DrawInfo) {
				((DrawInfo) m_drawInfo).paintComponent(g, shiftX, shiftY);
			} else {
				Enumeration en;
				DrawInfo	drawInfo;

				for (en = ((Vector) m_drawInfo).elements(); en.hasMoreElements(); ) {
					drawInfo = (DrawInfo) en.nextElement();
					drawInfo.paintComponent(g, shiftX, shiftY);
	}	}	}	}	

	public boolean isNearEdge(int x, int y, int threshold, int timeStamp) {

		if (timeStamp != -1 && timeStamp != this.timeStamp)
			return false;

		Vector pts = points(!isMarked(HIGHLIGHT_FLAG_MARK) && timeStamp != -1);

		if (pts == null || pts.size() < 2) {
		   return false;
		}
/*		System.out.println("RelationInstance:isNearEdge(x=" + x + ", y=" + y + ", threshold=" + threshold + ")");	// IJD
		for (int i = 0; i < pts.size(); ++i) {					// IJD
			System.out.println("pts[" + i + "]=" + pts.elementAt(i));
		}
*/
		return (findNearestSeg(x, y, threshold, pts) >= 0);
	}

	public void addAttribute(Attribute attr) 
	{
		processFirstOrderAttributes(attr, null);
	}

	public void assignAttributes(Attribute attr, Hashtable entities)
	{
		// Passed list of attributes

		for (; attr != null; attr = attr.next) {
			// For each attribute assignment, find the variable,
			// and assign the value.

			if (attr.avi == null) {
				MsgOut.println("Null attribute assignment");
			} else if (!processFirstOrderAttributes(attr, entities)) {
				Attribute attrVar = getAttribute(attr.id);

				if (attrVar == null) {
					MsgOut.println("Missing attribute '" + attr.toString() + "'");
				} else {
					if (attrVar.cloneOnAssign) {
						Attribute newAttr = (Attribute) attrVar.clone();
						replaceAttribute(newAttr);
					}
					attrVar.avi = attr.avi;
		}	}	}
	}

	public void writeRelation(PrintStream ps) throws IOException 
	{
		ps.print(qt(parentClass.getId()) + " " + qt(m_src.getId()) + " " + qt(m_dst.getId()) + "\n");
	}

	public void writeAttributes(PrintStream ps) throws IOException 
	{
		ByteArrayOutputStream	byteStream = new ByteArrayOutputStream();
		PrintStream				ps1        = new PrintStream(byteStream);
		String					s;

		super.writeAttributes(ps1, parentClass, false);

		s = byteStream.toString();
		if (s.length() != 0) {
			ps.print("(" + qt(parentClass.getId()) + " " + qt(m_src.getId()) + " " + qt(m_dst.getId()) + ") {\n" + s + "}\n");
		}
	}

	public boolean hasId(String id) 
	{
	   return false;
	}

	public boolean isSameRelation(RelationClass rc, EntityInstance src, EntityInstance dst) 
	{
		return (rc == parentClass && m_src == src && m_dst == dst); 
	}

	public EntityInstance getSrc() 
	{
		return m_src;
	}

	public EntityInstance getDst() 
	{
		return m_dst;
	}

	public EntityInstance getDrawSrc() 
	{
		return m_drawSrc;
	}

	public EntityInstance getCurrentSrc()
	{
		if (m_drawSrc != null) {
			return(m_drawSrc);
		}
		return(m_src);
	}

	public EntityInstance getDrawDst() 
	{
		return m_drawDst;
	}
	
	public EntityInstance getCurrentDst()
	{
		if (m_drawDst != null) {
			return(m_drawDst);
		}
		return(m_dst);
	}

	public RelationClass getRelationClass() 
	{
		return parentClass;
	}

	public void setHighlightFlag() 
	{
		if (!isMarked(HIGHLIGHT_FLAG_MARK)) {
			orMark(HIGHLIGHT_FLAG_MARK);
			repaint();
	}	}

	public boolean getHighlightFlag() 
	{
		return isMarked(HIGHLIGHT_FLAG_MARK);
	}

	public void clearHighlightFlag() 
	{
		if (isMarked(HIGHLIGHT_FLAG_MARK)) {
			nandMark(HIGHLIGHT_FLAG_MARK);
			validate();
	}	}

	public void setGroupAndHighlightFlag() 
	{
		if ((m_mark & (GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK)) != (GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK)) {
			orMark(GROUP_FLAG_MARK | HIGHLIGHT_FLAG_MARK);
			validate(); 
	}	}

	public void clearGroupFlag() 
	{
		if (isMarked(GROUP_FLAG_MARK)) {
			nandMark(GROUP_FLAG_MARK);
			validate();
	}	}

	public boolean getGroupFlag() 
	{
		return isMarked(GROUP_FLAG_MARK);
	}

	public boolean matches(RelationInstance other)
	{
		return(m_drawSrc == other.m_drawSrc && m_drawDst == other.m_drawDst);
	}

	public String toString() 
	{
		return m_drawSrc + (m_drawSrc == m_src ? "" : "{" + m_src + "}") + "->" + m_drawDst + (m_drawDst == m_dst ? "" : "{" + m_dst + "}");
	}

	public String toResultString()
	{
		return m_drawSrc.getLabel() + " " + getRelationClass().getLabel() + " " + m_drawDst.getLabel();
	}

	// MouseListener interface

	public void mouseClicked(MouseEvent e)
	{
	}

	public void mouseEntered(MouseEvent e)
	{
//		System.out.println("RelationInstance.MouseEntered " + getBounds());
		dg.getLs().setCursor(Cursor.HAND_CURSOR);
	}

	public void mouseExited(MouseEvent e)
	{
//		System.out.println("RelationInstance.MouseExited");
		dg.getLs().setCursor(Cursor.DEFAULT_CURSOR); 
	}

	public void mousePressed(MouseEvent ev)
	{
//		System.out.println("RelationInstance.mousePressed");
		
		dg.relationPressed(ev, this, ev.getX() + getX(), ev.getY() + getY());
	}

	public void mouseReleased(MouseEvent ev)
	{
//		System.out.println("RelationInstance.mousePressed");
		
		dg.relationReleased(ev, this, ev.getX() + getX(), ev.getY() + getY());
	}

	// MouseMotionListener interface

	public void mouseDragged(MouseEvent ev)
	{
		dg.relationDragged(ev, this, ev.getX() + getX(), ev.getY() + getY());
	}

	public void mouseMoved(MouseEvent ev)
	{
		if (this != m_infoShown) {
			String message;

			m_infoShown = this;
			if (m_src == m_drawSrc) {
				message = "";
			} else {
				message = Util.quoted(m_src.getLabel()) + "->";
			}
			message += Util.quoted(m_drawSrc.getLabel()) +  " " + getRelationClass().getLabel() + " " + Util.quoted(m_drawDst.getLabel());
			if (m_dst != m_drawDst) {
				message += "->" + Util.quoted(m_src.getLabel());
			}
//			message += " " + (getX() + ev.getX()) + "x" + (getY() + ev.getY());
			
			dg.getLs().showInfo(message);
		}
		dg.movedOverThing(ev, this, ev.getX() + getX(), ev.getY() + getY());
	}
}
